from enum import Enum, unique


@unique
class Types(Enum):
    """
        Tipos de elementos soportados en los microservicios
    """
    WORKER = 1
    FORK = 2


@unique
class TypesActions(Enum):
    CREATE = 1
    GET = 2
    DELETE = 3
    LIST = 4
    LISTENER = 5
    UPDATE = 6
    ERRORS = 7
    FORKS = 8


@unique
class Actions(Enum):
    """
        Acciones que se ejecutan desde un servicio
    """
    CREATE = 1
    GET = 2
    DELETE = 3
    LIST = 4
    LISTENER = 5
    UPDATE = 6


class HttpError(Enum):
    BadRequest = 400
    NotAuthenticated = 401
    PaymentError = 402
    Forbidden = 403
    NotFound = 404
    MethodNotAllowed = 405
    NotAcceptable = 406
    Timeout = 408
    Conflict = 409
    LengthRequired = 411
    Unprocessable = 422
    TooManyRequests = 429
    GeneralError = 500
    NotImplemented = 501
    BadGateway = 502
    Unavailable = 503



import asyncio
import base64
import json
import sys
from nats.aio.client import Client as NATS
from nats.aio.errors import ErrNoServers
from .loggs import Loggs


class NatsBase:

    # Instancia de Nats Client
    __nc = NATS()

    # Logs de la aplicacion
    logs = Loggs('Service')

    # Nombre del agente
    __Name = ''

    # Hots De Nats
    __Nats_Host = ''

    # Indicar si es un cola de procesos
    __isQueue = False

    # Nombre del cliente
    __name = None

    # Loop de la aplicacion
    loop = None

    # Init
    def __init__(self, agentWorkerName, NatsHosts, name=None, Queue=True):
        self.__Name = agentWorkerName
        self.__Nats_Host = NatsHosts
        self.__isQueue = Queue
        self.__name = name

    def async_manager(self, coro):
        """
            Metodo para poder hacer el llamado de la funcion Async
            sin necesidad de usar await.

            @params coro Metodo async a llamar
        """
        try:
            coro.send(None)
        except StopIteration as stop:
            return stop.value

    async def run(self):
        # Servidores
        servers = []

        # Dar el formato correcto
        for ntServer in self.__Nats_Host:
            servers.append('nats://{}'.format(ntServer))

        # Setup pool of servers from a NATS cluster.
        options = {
            "servers": servers,
            "io_loop": asyncio.get_event_loop(),
            "name": self.__name,
            'dont_randomize': True,
            'max_reconnect_attempts': 5,
            'reconnect_time_wait': 2,
            "disconnected_cb": self.disconnected_cb,
            "reconnected_cb": self.reconnected_cb,
            "error_cb": self.error_cb,
            "closed_cb": self.closed_cb
        }

        # Connect
        try:
            await self.__nc.connect(**options)
        except ErrNoServers as e:
            # Could not connect to any server in the cluster.
            print(e)
            sys.exit(-1)

        # Validar que se haya establecido una conexion
        if self.__nc.is_connected:
            # Conectar y escuchar el evento del metodo unico
            if self.__isQueue:
                await self.__nc.subscribe(
                    self.__Name,
                    queue=self.__Name,
                    cb=self.subscribe_handler
                )
            else:
                await self.__nc.subscribe(
                    self.__Name,
                    cb=self.subscribe_handler
                )
            self.logs.info(
                'Conexion establecida con el servidor de NATS: {}'.format(
                    self.__Nats_Host
                )
            )

            # Notificar que se encuentra conectado
            self.logs.info("Escuchando el topico {} en NATS {}".format(
                self.__Name, ','.join(self.__Nats_Host)))

        err = self.__nc.last_error
        if err is not None:
            self.logs.info("Ocurrio un error: {}".format(err))

    @asyncio.coroutine
    async def disconnected_cb(self):
        self.logs.warning(
            "Se realizo una desconexion de NATS: {}".format(self.__Nats_Host))

    @asyncio.coroutine
    async def reconnected_cb(self):
        self.logs.info("Reconectando a {url}".format(
            url=self.__nc.connected_url.netloc))

    @asyncio.coroutine
    async def error_cb(self, e):
        self.logs.error("[NATS] Ocurrio el siguiente Error: {}".format(e))

    @asyncio.coroutine
    async def closed_cb(self):
        self.logs.info("Se cerro la conexion al servidor de NATS")

    async def subscribe_handler(self, msg):
        # Convertir en un objeto la peticion recivida
        DATA = self.__b64_to_json(msg.data)
        # Validar si es un Queue
        if self.__isQueue and DATA.get('uuid', None) is None:
            if DATA['headers']['Asynchronous']:
                DATA.update({"uuid": None})
            else:
                # Asignar el ID para que le responda
                DATA.update({"uuid": msg.reply})

        # Pasar los datos
        """ try:
            # _thread.start_new_thread(self._message, (DATA))
            # self._message(msg.value)
            hilo = threading.Thread(target=self._message, args=(DATA,))
            hilo.start()
        except Exception as e:
            print(e) """
        # _thread.start_new_thread(self._message, (DATA,))
        self._message(DATA)

    def __json_to_b64(self, json_in):
        """
            Metodo que pasa un objecto a el formato necesario
            para su comunicacion
        """
        JSON_SER = None
        try:
            JSON_SER = json.dumps(json_in)
        except Exception:
            self.logs.error(
                'Ocurrio un error al momento de serializar el JSON')

        return base64.b64encode(str.encode(JSON_SER))

    def __b64_to_json(self, encoded):
        """
            Metodo que conviernte un base64 a dict
        """
        decoded = base64.b64decode(encoded)
        return json.loads(decoded.decode('utf-8'))

    def _message(self, msg):
        """
            Metodo para el procesamiento de mensajes de
            Nats.
        """
        pass

    async def ___send_async(self, reply, msj, idTransaction):
        await self.__nc.publish(reply, msj)

    def _send(self, reply, msj, idTransaction=""):
        """
            Metodo para el envio de datos a Nats

            @params reply Elemento al que se le responde
            @params msj Mensaje que enviara
            @params idTransaction Id de la transaccion
        """
        self.async_manager(self.___send_async(
            reply, self.__json_to_b64(msj), idTransaction))

    def isQueue(self):
        """
            Metodo para conocer si un queue worker
        """
        return self.__isQueue

    def validateConnection(self):
        return self.__nc.is_connected



from .ErrorMs import ErrorMs


class BadRequest(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 400, headers)


class NotAuthenticated(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 401, headers)


class PaymentError(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 402, headers)


class Forbidden(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 403, headers)


class NotFound(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 404, headers)


class MethodNotAllowed(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 405, headers)


class NotAcceptable(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 406, headers)


class Timeout(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 408, headers)


class Conflict(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 409, headers)


class LengthRequired(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 411, headers)


class Unprocessable(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 422, headers)


class TooManyRequests(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 429, headers)


class GeneralError(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 500, headers)


class NotImplemented(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 501, headers)


class BadGateway(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 502, headers)


class Unavailable(ErrorMs):
    def __init__(self, code, headers={}, message=""):
        super().__init__(code, message, 503, headers)



from .Base import Base
from .ErrorMs import ErrorMs
from .HttpResponse import HttpResponse


class Worker(Base):
    """
        Clase que contiene la logica para el manejo de los
        worker.
    """

    def process(self, span, request, fnc, params=[]):
        """
            Metodo que se encarga de procesar, la accion
            que se asign al evento del microservicio.

            @params resquest peticion de datos
            @params fnc Funcion a ejecutar
        """
        try:
            # Parametros
            paramsInject = self.getParams(span, request, params)

            # Ejecutar la funciona final
            RESP = fnc(*paramsInject)
            # Regresar la respuesta con el formato correcto
            return self.formatResponse(request, RESP)
        except ErrorMs as err:
            # Retornar el error
            return self.formatResponse(
                request,
                # Generar formato
                HttpResponse(None, err.http_code, headers=err.headers),
                err.errorCode
            )



import os
import socket
from bottle import Bottle


class HttpServer:
    """
        Clase para el manejo de peticiones utilizando un servior HTTP en lugar
        de las colas de mensajes.
    """

    def __init__(self, topic, app_name, version, service_name):
        # Variables de configuracion
        self._routes = []

        # Configuracion del servidor
        self._host = '0.0.0.0'
        self._port = os.environ.get('HTTP_PORT', 8000)

        # Configuracion del servicio
        self.__hostname = socket.gethostname()
        self._topic = topic
        self._app_name = app_name
        self._version = version
        self._service_name = service_name

        # Configuracion del bottle
        self._app = Bottle()
        self.__init_list_routes()  # Generar las URL
        self.__init_router()  # Iniciar las rutas del servidor

    def __init_list_routes(self):
        """
            Metodo para iniciar la lista de las rutas
        """
        self._routes = [
            "/{}/api/{}/{}".format(self._app_name,
                                   self._version, self._service_name),
            "/{}/api/{}/{}/<id>".format(self._app_name,
                                        self._version, self._service_name),
            "/api/{}/{}".format(self._version, self._service_name),
            "/api/{}/{}/<id>".format(self._version, self._service_name)
        ]

    def __init_router(self):
        """
            Metodo para realizar la inicializacion de las rutas
            para el servidor.
        """
        for route in self._routes:
            self._app.route(
                route,
                ['GET', 'POST', 'UPDATE', 'DELETE', 'OPTION'],
                self._process_handler
            )

    def _process_handler(self, id=None):
        """
            Handler del servicio http
        """
        pass

    def start(self):
        """
            Metodo para iniciar el servidor HTTP
        """
        self._app.run(host=self._host, port=self._port)



from pygments import highlight, lexers, formatters


def ColorsJson(formatted_json):
    """
    """
    # return highlight(unicode(formatted_json, 'UTF-8'), lexers.javascript() , formatters.TerminalFormatter())
    pass



import sys
import os
from .loggsWrapper import LoggerWrapper


def Loggs(name, logsAll=True):
    """
        Metodo para generar un elemento de logger
        @params name Nombre del modulo/libreria
    """

    # Validar si proviene de un ambiente productivo.
    if not os.environ.get('LOGS_FLUENT', None) is None:
        return LoggerWrapper(name)
    else:
        import coloredlogs
        import logging

        logger = logging.getLogger(name)

        # solo cuando se indique se mostraran todos los logs
        if os.environ.get('PRODUCTION', False):
            from logbook import Logger, StreamHandler
            import sys
            StreamHandler(sys.stdout).push_application()
            return Logger(name)
        elif logsAll:
            coloredlogs.install()  # logger=logger
        else:
            coloredlogs.install()
        return logger



import time
import math
import datetime
from .Base import Base
from .Util import getItemBy, verbAction
from .HttpResponse import HttpResponse


class Fork(Base):

    # Configuracion de los forks
    ___conf_forks = {}

    # Forks
    ___forks = []

    # Primer bifurcacion
    ___isFirstFork = False

    def process(self, span, data, fnc, params=[]):
        """
            Metodo que se encarga de realizar le procesamiento de
            los mensajes pasados por Microservices
        """

        # Posicion del actual Fork
        worker = self.nextFork(getItemBy('metadata.owner', data))
        # Agregar el tiempo actual
        data['metadata']['time'] = datetime.datetime.utcnow().isoformat()
        # Indicar tipo
        data['metadata']['mtype'] = 'input'
        # Validar si tiene asignado la metadata original
        if getItemBy('metadata.metadata_inicial', data) is None:
            data['metadata']['metadata_inicial'] = data['metadata'].copy()

        # Validar si tiene almacenada los datos originales
        if getItemBy('metadata.data_inicial', data) is None:
            data['metadata']['data_inicial'] = data['data'].copy()

        owner = getItemBy('metadata.owner', data)

        # Validar si es consecutivo o primer servicio
        if self.Topico != owner or len(self.___forks) < 1:
            # Recuperar el siguiente
            nextFork = self.nextFork(getItemBy('metadata.owner', data))

            # Validar si ocurrio un error
            isError = self.isError(data)

            # Validar si aun se tienen elementos que procesar (Forks)
            if nextFork is None or isError:
                data['metadata']['bifurcacion'] = False
            else:
                data['metadata']['bifurcacion'] = True

            if isError:
                # Retornar el mensaje correcto
                return self.formatResponse(
                    data,
                    # Enviar el resultado correcto
                    self.get_http_data(data['metadata']),
                    self.getErrorCode(data),
                    self.getErrorMeta(data)
                )
            elif nextFork is None:
                # Parametros
                paramsInject = self.getParams(span, data, params)

                # Ejecutar la funciona final
                RESP = fnc(*paramsInject)

                # Regresar la respuesta con el formato correcto
                return self.formatResponse(data, RESP)
            else:
                row = self.getNextTopic(nextFork)
                conf = self.___conf_forks.get(row.get('current'))
                method = data['metadata']['method']

                data['metadata']['original_method'] = method
                data['metadata']['method'] = verbAction(conf.get('action'))

                resp = None

                r1 = getItemBy('response', data)
                r2 = getItemBy('response.data', data)
                r3 = getItemBy('response.data.response', data)

                if r1 and r2 and r3:
                    resp = getItemBy('response.data.response', data)

                data['data'] = self.getData(
                    getItemBy('metadata.data_inicial', data),
                    conf,
                    resp,
                )

                data = self.changeMetadata(data, row.get('current'))

                return self.formatResponse(data, None, 0)

        else:
            # Recuperar los datos requeridos del siguiente Topico
            row = self.getNextTopic(worker)

            # Recuperar la configuracion
            conf = self.___conf_forks.get(row.get('current'), None)

            # Indicar que es una bifurcacion
            data['metadata']['bifurcacion'] = True

            # Agregar el metodo que corresponde y se
            # guarda una copia para usarla despues
            data['metadata']['original_method'] = data['metadata']['method']
            data['metadata']['method'] = verbAction(conf.get('action'))

            # Recuperar la configuracion
            data['data'] = self.getData(
                data['data'],
                conf,
                None,
            )

            # Aplicar los cambios
            data = self.changeMetadata(data, row.get('current'))

            # Retornar la respuesta
            return self.formatResponse(data, None, 0)

    def confForks(self, conf):
        """
            Metodo para registrar la configuracion del metodo
            que se ejecutara.
        """
        if conf is not None:
            # Asignar los paso a repetir.
            for item in conf:
                # Configuracion
                conf_item = item.get('conf')
                if not self.existService(conf_item):
                    Version = conf_item.get('Version')
                    # Validar el formato del ID
                    if conf_item.get('App') == 'appcoppel' and Version == 'v1':
                        self.___conf_forks.update(
                            {conf_item.get('Name'): item})
                    else:
                        # Id basado en la version 2
                        id_unique = '{}_{}_{}'.format(
                            conf_item.get('App'),
                            conf_item.get('Version'),
                            conf_item.get('Name'),
                        )
                        self.___conf_forks.update({id_unique: item})

                    # Coleccion de forks para la accion.
                    self.___forks.append(item.get('conf', None))

    def existService(self, conf):
        """
            Metodo para validar si nu servicio ya existe
            previamente registrado.
        """
        for item in self.___forks:
            APP = item.get('App')
            NAME = item.get('Name')
            if APP == conf.get('App') and NAME == conf.get('Name'):
                return True

        return False

    def getData(self, dataDefault, conf, response):
        """
            Metodo para recuperar el data que se ejecutara

            @param dataDefault Data por default
            @param conf Configuracion del fork
            @param response Datos del servicio previo
        """
        if conf.get('fnc_call', None):
            # Recuperar la funcion
            fnc_call = conf.get('fnc_call')
            # Ejecutar la funcion los parametros de
            # los datos nuevos y originales
            resp = fnc_call(dataDefault, response)
            # Validar que sea un diccionario
            if isinstance(resp, dict):
                return resp
            else:
                # Wrap para no ocacionar error
                return {"value": resp}
        elif conf.get('data', None):
            return conf.get('data')
        else:
            return dataDefault

    def getNextTopic(self, worker):
        """
            Metodo para crear la estructura para la publicacion del mensaje.
            @param worker Configuracion del worker
        """
        # Generar el elemento actual
        current = '{}_{}_{}'.format(
            worker.get('App'),
            worker.get('Version'),
            worker.get('Name'),
        )
        if worker.get('Version') == 'v1' and worker.get('App') == 'appcoppel':
            current = worker.get('Name')

        return {'current': current, 'name': worker.get('Name')}

    def nextFork(self, nameFork):
        """
            Metodo para buscar la posicion del siguiente Fork a ejecutar

            @params nameFork Nombre que se busca
        """
        # Ver el tamaño las bifurcaciones asignadas
        size = len(self.___forks)
        if size == 0:
            return None

        # Ver si no corresponde al propio topico
        if nameFork == self.Topico:
            self.___isFirstFork = True
            return self.___forks[0]

        # Buscar la posicion del elemento actual
        for index in range(0, size):
            obj = self.___forks[index]
            rpTopic = self.getNextTopic(obj)
            # Validar si es el que corresponde
            if rpTopic.get('current') == nameFork:
                if (index + 1) < size:
                    return self.___forks[index+1]
                else:
                    return None

        # Valor por default
        self.___isFirstFork = True
        return self.___forks[0]

    def get_http_data(self, meta, data=None):
        """
            Metodo para recuperar los datos HTTP de una worker previo
        """
        if "http" in meta:
            HTTP = meta["http"]
            return HttpResponse(
                data,
                HTTP["code"],
                HTTP["message"],
                HTTP["headers"]
            )
        else:
            return data

    def isError(self, data):
        """
            Metodo para asergurar si ocurrio un error al
            momento de consumir el servicio
        """
        r = getItemBy('response', data)
        rp = getItemBy('response.data.response', data)
        ec = False if getItemBy(
            'response.data.response.errorCode', data) is None else True
        return r and rp and ec

    def getErrorCode(self, data):
        """
            Metodo para recuperar el error code
        """
        # Recuperar el codigo de error
        errorCode = getItemBy('response.data.response.errorCode', data)
        # Validar que se proporcionara un codigo de error.
        if errorCode is None:
            # Valor por default
            return -99
        else:
            return errorCode

    def getErrorMeta(self, data):
        """
        Metodo para recuperar la meta data de los errore.

        Arguments:
            data object -- Datos de la repuesta
        """
        # Recuperar todo el objeto de error.
        errorCode = getItemBy('response.data.response.errorCode', data)

        # Validar que sea un error y no un dato correcto.
        if errorCode is None:
            return None
        else:
            return getItemBy('response.data.response', data)

    def changeMetadata(self, data, current):
        """
            Metodo apra aplicar los cambios sobre la metadata
        """
        data['metadata']['id_operacion'] = math.floor(time.time()) * 433333333
        data['metadata']['uowner'] = getItemBy('metadata.uworker', data)
        data['metadata']['worker'] = current  # Worker que procesara
        data['metadata']['uworker'] = '{}_{}'.format(
            getItemBy('metadata.worker', data),
            getItemBy('metadata.id_operacion', data),
        )
        data['metadata']['owner'] = self.Topico
        # Regresar los nuevos datos
        return data



"""
    Modulo para la implementacion de los decoradores,
    necesarios para el uso de reflec data, al momento
    de registrar las acciones de los microservicios.
"""

import inspect
import sys
from .types import TypesActions
from .loggs import Loggs

logs = Loggs('Core')


def validForm(forkConf):
    """
        Metodo para realizar la validacion del diccionario
        de configuracion de implementacion un Worker.
    """
    if len(forkConf) > 0:
        for conf in forkConf:

            invalid = False

            # Validar que cuente con los elementos minimos
            if conf.get('fork', None) is None:
                invalid = True
            """ elif conf.get('fnc', None) is None:
                if conf.get('data', None) is None:
                    invalid = True """

            if invalid:
                logs.error(
                    'La configuracion para la bifurcacion no es correcta')
                sys.exit(-1)


def validParam(fn):
    """
        Metodo para validar que se proporcionaran
        los parametros minimos para la funciones
        que interactuan con un registro expecifico.
    """
    if len(inspect.getargspec(fn)[0]) < 4:
        logs.error(
            'Lo metodos GET, DELETE, UPDATE requieren 3 parametros')
        sys.exit(-1)


def asignDec(fn, typeDec):
    """
        Metodo para agregar el decorador y los atributos
        necesarios para el reflect data
    """
    # Decorador de la funcion
    def decorated(*args, **kwargs):
        return fn(*args, **kwargs)
    # Asignar los atributos
    if typeDec == TypesActions.ERRORS or inspect.isfunction(fn):
        setattr(decorated, '__MICROSERVICE_ACTION__', True)
        setattr(decorated, '__TYPE__', typeDec)
        setattr(decorated, '__FUNC_NAME__', fn.__name__)
    return decorated


def Define(opt):
    """
        Metodo que agrega la informacion
        necesaria para iniciar la clase del
        microservicio.
    """
    def dec(cls):
        try:
            cls(opt)
        except KeyboardInterrupt:
            cls.logs.info('Keyboard Interrupt')
        except RuntimeError:
            cls.logs.info('Runtime Error')
        return cls
    return dec


def List(forks=[], ctx=[]):
    """
        Decorador para las acciones de tipo
        LIST (GET a la raiz del servicio)
    """
    validForm(forks)

    def x(fn):
        # Decorador original
        afn = asignDec(fn, TypesActions.LIST)
        # Asginar los datos
        if len(forks) > 0:
            setattr(afn, '__CONF_FORKS__', forks)
        # Configuracion de parametros
        if len(ctx) > 0:
            setattr(afn, '__CONF_PARAMS__', ctx)
        return afn
    return x


def Get(forks=[], ctx=[]):
    """
        Decorador de la accion cuando se solicitan
        los datos de un elemento especifico
    """
    validForm(forks)

    def x(fn):
        # Decorador original
        afn = asignDec(fn, TypesActions.GET)
        # Asginar los datos
        if len(forks) > 0:
            setattr(afn, '__CONF_FORKS__', forks)
        # Configuracion de parametros
        if len(ctx) > 0:
            setattr(afn, '__CONF_PARAMS__', ctx)
        # Retornar el decorador
        return afn
    return x


def Create(forks=[], ctx=[]):
    """
        Decorador de accion de creacion de un
        elemento en el servicio
    """
    validForm(forks)

    def x(fn):
        # Decorador original
        afn = asignDec(fn, TypesActions.CREATE)
        # Asginar los datos
        if len(forks) > 0:
            setattr(afn, '__CONF_FORKS__', forks)
        # Configuracion de parametros
        if len(ctx) > 0:
            setattr(afn, '__CONF_PARAMS__', ctx)
        # Retornar el decorador
        return afn
    return x


def Update(forks=[], ctx=[]):
    """
        Decorador de actualizacion de un
        servicio en especfico
    """
    validForm(forks)

    def x(fn):
        # Decorador original
        afn = asignDec(fn, TypesActions.UPDATE)
        # Asginar los datos
        if len(forks) > 0:
            setattr(afn, '__CONF_FORKS__', forks)
        # Configuracion de parametros
        if len(ctx) > 0:
            setattr(afn, '__CONF_PARAMS__', ctx)
        # Retornar el decorador
        return afn
    return x


def Delete(forks=[], ctx=[]):
    """
        Decorador de accion de eliminacion
        de un elemento especfico
    """
    validForm(forks)

    def x(fn):
        # Decorador original
        afn = asignDec(fn, TypesActions.DELETE)
        # Asginar los datos
        if len(forks) > 0:
            setattr(afn, '__CONF_FORKS__', forks)
        # Configuracion de parametros
        if len(ctx) > 0:
            setattr(afn, '__CONF_PARAMS__', ctx)
        # Retornar el decorador
        return afn
    return x


def Listener(forks=[], ctx=[]):
    """
        Decorador Generico
    """
    validForm(forks)

    def x(fn):
        # Retornar el decorador
        afn = asignDec(fn, TypesActions.LISTENER)
        # Asginar los datos
        if len(forks) > 0:
            setattr(afn, '__CONF_FORKS__', forks)
        # Configuracion de parametros
        if len(ctx) > 0:
            setattr(afn, '__CONF_PARAMS__', ctx)
        return afn
    return x


def Errors(fn):
    """
        Decorador para asignacion de los errores
    """
    # Retornar el decorador
    return asignDec(fn, TypesActions.ERRORS)


def Forks(fn):
    """
        Decorador para asignacion de los Forks
    """
    # Retornar el decorador
    return asignDec(fn, TypesActions.FORKS)



class HttpResponse:
    """
        Clase base para respuestas dentro de un servicio HTTP
    """
    # Datos que seran enviados
    data = {}

    # Codigo que sera enviado
    code = 200

    # Mensaje que se quiere pasar
    message = ""

    # Headers
    headers = {}

    def __init__(self, data, code, message="", headers={}):
        """
            Constructor de la clase para la respuesta de HTTP
        """
        self.data = data
        self.code = code
        self.message = message
        self.headers = headers



from .microservices import Microservices
from .ms_base import KafkaBase
from .loggs import Loggs
from .types import Types, TypesActions, Actions, HttpError
from .options import Options
from .ErrorMs import ErrorMs
from .version_framework import version
from .HttpResponse import HttpResponse
from .HttpErrors import BadRequest

name = 'MsCoppel'

__all__ = [
    'Microservices',
    'KafkaBase',
    'Loggs',
    'Types',
    'Options',
    'MsManager',
    'TypesActions',
    'ErrorMs',
    'Actions',
    'HttpResponse',
    'HttpError',
    'BadRequest',
]

__version__ = version



from .types import Actions


class Struct:
    def __init__(self, **entries):
        self.__dict__.update(entries)


def Merge(dict1, dict2):
    return(dict2.update(dict1))


def getItemBy(anotation, data, default=None):
    """
        Metodo para acceder en anotacion de objeto a las propiedades
        de un diccionario.
    """
    sep = anotation.split('.')
    if len(sep) > 0:
        current = data
        for item in sep:
            if current.get(item, None) is None:
                return default
            else:
                current = current.get(item, default)
        # Regresar el resultado del ciclo
        return current
    else:
        # No cuenta con un solo elemento
        return default


def verbAction(action):
    """
        Metodo para recuperar el verbo de una
        accion para el envio de parametro.
    """
    if action == Actions.CREATE:
        return 'POST'
    elif action == Actions.DELETE:
        return 'DELETE'
    elif action == Actions.GET:
        return 'GET'
    elif action == Actions.LIST:
        return 'GET'
    elif action == Actions.UPDATE:
        return 'PUT'
    return 'GET'


def validForksConf(conf):
    """
        Metodo para validar la configuracion
        de los Forks tiene el formato
        correspondiente.
    """
    error = None

    # Validar que sea un diccionario
    if not isinstance(conf, dict):
        error = 'No se prporciono una configuracion globar de Forks Correcta'
        return False, error

    # Recorrer la configuracion para validar los metodos
    for name, confservice in conf.items():
        if not isinstance(confservice, dict):
            error = 'La configuracion del servicio {} no es correcta'.format(
                name)
            return False, error

        # Validar que cuente con los elementos basicos
        if confservice.get('App', None) is None:
            error = 'Se requiere la configuracion "APP" para el servicio "{}"'.format(
                name)
        elif confservice.get('Name', None) is None:
            error = 'Se requiere la configuracion "Name" para el servicio "{}"'.format(
                name)
        elif confservice.get('Version', None) is None:
            error = 'Se requiere la configuracion "Version" para el servicio "{}"'.format(
                name)

        if not error is None:
            return False, error

    return True, error



class Error(Exception):
    """Clase base para los errres"""
    pass


class ErrorMs(Error):
    """
        Clase para la definicion de errores dentro de
        los microservicios
    """
    errorCode = 0
    message = ""
    headers = {}
    http_code = 0

    def __init__(self, error_code=-99, message='',
                 http_status_code=None,  headers={}):

        self.errorCode = error_code
        self.message = message
        self.headers = headers
        self.http_code = http_status_code
        super().__init__(self.message)



import base64
import json
import os
import sys
import _thread
from kafka import KafkaConsumer, KafkaProducer
from .loggs import Loggs


class KafkaBase:
    """
        Base del microservicio que se aplica para
        la comunicacion con kafka
    """
    # Topico de conexion
    __TOPIC = ''
    # Direccion de kafka
    __KAFKAHOSTS = ''
    # Instancia del consumer
    __CONSUMER = None
    # Instancia del producer
    __PRODUCER = None
    # Logs de la aplicacion
    logs = Loggs('Service')

    def __init__(self, topic, kafkaHosts, name=None):
        self.logs.info('Iniciando la conexion los servidores de KAFKA')
        # Asignar el topico de conexion
        self.__TOPIC = topic
        # Asginar el hosts de kafka
        self.__KAFKAHOSTS = kafkaHosts
        # Conectar a Kafka Producer
        self.__connectProducer()
        # Conectar a Kafka Consumer
        self.__connectConsumer()

    def __connectConsumer(self):
        """
            Metodo para realizar la conexion al cosumer de kafka.
        """
        topics = [self.__TOPIC, "validateConnection"]

        CONSUMER_GROUP = os.environ.get('CONSUMER_GROUP', self.__TOPIC)

        # Conexion a kafka
        try:
            self.__CONSUMER = KafkaConsumer(
                *topics,
                group_id=CONSUMER_GROUP,
                auto_offset_reset='earliest',
                enable_auto_commit=True,
                bootstrap_servers=self.__KAFKAHOSTS,
                value_deserializer=self.__b64_to_json
            )
        except Exception as e:
            self.logs.error('Ocurrio un error al conectar con Kafka')
            self.logs.error(e)
            sys.exit(-1)

        # Notificar que se encuentra conectado
        self.logs.info("Escuchando el topico {} en el Kafka {}".format(
            self.__TOPIC, ','.join(self.__KAFKAHOSTS)))

        # Escuchar todos los posibles eventos
        for msg in self.__CONSUMER:
            try:
                # Solo procesar los mensajes del topico
                if msg.topic == self.__TOPIC:
                    _thread.start_new_thread(self._message, (msg.value,))
                    # self._message(msg.value)
            except Exception as e:
                print(e)

    def _message(self, msg):
        """
            Metodo para el procesamiento de mensajes de
            kafka.
        """
        pass

    def __connectProducer(self):
        """
            Metodo para realizar la conexion con el producer
            de kafka.
        """
        try:
            self.__PRODUCER = KafkaProducer(
                bootstrap_servers=self.__KAFKAHOSTS,
                value_serializer=self.__json_to_b64
            )
            self.logs.info('Conectado a kafka para enviar mensajes')
        except:
            self.logs.error('Ocurrio un error al conectar con el producer')
            sys.exit(-1)

    def __json_to_b64(self, json_in):
        """
            Metodo que pasa un objecto a el formato necesario
            para su comunicacion
        """
        return base64.b64encode(str.encode(json.dumps(json_in)))

    def __b64_to_json(self, encoded):
        """
            Metodo que conviernte un base64 a dict
        """
        decoded = base64.b64decode(encoded)
        return json.loads(decoded.decode('utf-8'))

    def _send(self, topico, msj, idTransaction):
        """
            Metodo para el envio de datos a Kafka

            @params topico Topico en el que se publica
            @params msj Mensaje que enviara
            @params idTransaction Id de la transaccion
        """
        try:
            self.__PRODUCER.send(topico, key=str.encode(
                str(idTransaction)), value=msj)
        except Exception as e:
            self.logs.error(e)
            sys.exit(-1)

    def validateConnection(self):
        try:
            self.__CONSUMER.topics()
            self.__PRODUCER.send('validateConnection', '')
            return True
        except:
            return False



from abc import abstractmethod, ABC
from .Util import getItemBy


class Base(ABC):
    """
        Clase base para la implementacion de los tipos
        Worker y Fork
    """

    Topico = ''

    def __init__(self, Topico):
        self.Topico = Topico

    @abstractmethod
    def process(self, span, request, fnc, params=[]):
        """
            Metodo que se encargara de procesar los mensajes
            que se enviaran al microservicio.
        """
        return 0

    def confForks(self, conf):
        """
            Metodo para registrar la configuracion del metodo
            que se ejecutara.
        """
        pass

    def formatResponse(self, original, resp, errorCode=0, errorWorker=None):
        """
            Metodo que se encarga de darle formato a la
            salida del procesamiento de la accion.
        """

        errorCodeBk = None

        try:
            errorCodeBk = errorCode
            # Prevenir el paso de string como error
            errorCode = int(errorCode)
        except Exception:
            print(
                'Ocurrio une error al realizar el CAST del error: {}'.format(
                    errorCodeBk
                )
            )
            errorCode = -99

        # Regresar la respuesta con el formato correcto
        if errorWorker is not None:
            pass

        resp = {
            "errorCode": errorCode,
            "response": {
                "data": original.get("data", None),
                "headers": original.get("headers", {}),
                "metadata": original.get("metadata", {}),
                "response": {
                    "data": {
                        "response": resp
                    },
                    "meta": {
                        "id_transaction": original["metadata"]["id_transaction"],
                        "status": 'ERROR' if errorCode < 0 or errorCode > 0 else 'SUCCESS'
                    }
                },
                "uuid": original.get("uuid", None),
            },
            "errorWorker": errorWorker
        }

        # Retornar los datos
        return resp

    def getParams(self, span, data, paramList):
        """
            Metodo para recuperar los datos segun la configuracion
            de los parametros.
        """
        # Parametros
        paramsInject = []

        # Heders
        HEADERS = data["headers"]

        # Auth
        AUTH = HEADERS.get("Authorizacion", None)

        # logs
        log = {'event': 'input'}

        # En caso de que se indique con otro nombre
        if AUTH is None:
            AUTH = HEADERS.get("Authorization", None)

        # Seleccion de parametros
        for i in paramList:
            if i == 'data':
                paramsInject.append(data.get('data'))
                log['data'] = data.get('data')
            elif i == 'filters':
                data_params = self.__transformFilter(
                    getItemBy('metadata.filters', data))
                data_params = {} if data_params is None else data_params
                paramsInject.append(data_params)
                log['filters'] = data_params
            elif i == 'authorization':
                paramsInject.append(AUTH)
                log['authorizacion'] = AUTH
            elif i == 'authorizacion':
                paramsInject.append(AUTH)
                log['authorizacion'] = AUTH
            elif i == 'uuid':
                paramsInject.append(getItemBy('metadata.uuid', data))
                log['uuid'] = getItemBy('metadata.uuid', data)
            elif i == 'response':
                if 'response' in data and 'data' in data['response'] and 'response' in data['response']['data']:
                    paramsInject.append(
                        getItemBy('response.data.response', data))
                    log['response'] = getItemBy('response.data.response', data)
                else:
                    paramsInject.append(data.get('data'))
                    log['response'] = data.get('data')

        span.log_kv(log)

        return paramsInject

    def __transformFilter(self, data):
        """
            Metodo para dar un forma coherente a los filtros
        """
        resp = {}
        # Validar que existan datos
        if data is None:
            return data

        for k in data:
            l = data.get(k)
            if len(l) == 1:
                if len(l[0].split(',')) > 1:
                    resp.update({k: l[0].split(',')})
                else:
                    resp.update({k: l[0]})
            else:
                resp.update({k: l})

        return resp



import os
import socket
import time
import datetime
from fluent import asyncsender as sender

class LoggerWrapper(object):

    name = 'FLUENT_LOGS'

    __name = None
    __APP = ''
    __SERVICE = ''
    __VERSION = ''
    __BROKERS = ''
    __FRAMEWORK_VERSION = ''
    __HOSTNAME = ''
    __PRODUCTION = False

    __LEVELS = {
        'FATAL': 60,
        'ERROR': 50,
        'WARNING': 40,
        'INFO': 30,
        'DEBUG': 20,
        'TRACE': 10,
    }

    __logger = None

    def __init__(self, name):
        self.__name = name
        # Conexion con Fluentd
        self.__logger = sender.FluentSender(
            'coppelio.logs',
            host=os.environ.get('FLUENTD_HOST', 'localhost'),
            port=os.environ.get('FLUENTD_PORT', 24224),
            queue_circular=True
        )
        # Asignar el hostname
        self.__HOSTNAME = socket.gethostname()
        # Production
        self.__PRODUCTION = False if os.environ.get('PRODUCTION', None) is None else True

    def set_conf(self, app, service, version, brokers, frmversion):
        """
            Metodo para realizar la configuraciones extra del servicio
            para su procesamiento en Fluentd.
        """
        self.__APP = app
        self.__SERVICE = service
        self.__VERSION = version
        self.__BROKERS = brokers
        self.__FRAMEWORK_VERSION = frmversion

    def trace(self, message, data={}):
        self._log('TRACE', message, data)

    def debug(self, message, data={}):
        self._log('DEBUG', message, data)

    def info(self, message, data={}):
        self._log('INFO', message, data)

    def warn(self, message, data={}):
        self._log('WARNING', message, data)

    def warning(self, message, data={}):
        return self.warn(message, data)

    def fatal(self, message, data={}):
        self._log('FATAL', message, data)

    def error(self, message, data={}):
        self._log('ERROR', message, data)

    def log(self, level, message, data={}):
        self._log(level, message, data)

    def _log(self, level, message='', data={}):
        # Recuperar el level
        _level = self.__LEVELS.get(level, 10)
        # Logss
        log = {
            '@timestanp': datetime.datetime.now().isoformat(),
            'log_type': 'app',
            'level': _level,
            'type': level,
            'production': self.__PRODUCTION,
            'app': self.__APP,
            'service': self.__SERVICE,
            'version': self.__VERSION,
            'brokers': self.__BROKERS,
            'framework_version': self.__FRAMEWORK_VERSION,
            'hostname': self.__HOSTNAME,
            'message': message,
            'data': data,
        }
        # Publicar el log
        self.__logger.emit_with_time('app', int(time.time()), log)



import setuptools
from MsCoppel import version

with open("README.md", "r") as fh:
    long_description = ''  # fh.read()

setuptools.setup(
    name="MsCoppel",
    version=version,
    author="Enoi Barrera Guzman",
    author_email="zafiro3000x@gmail.com",
    description="Libreria para microservicios basados en mensajes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3"
    ],
    install_requires=[
        'kafka-python==2.0.1',
        'Logbook',
        'asyncio-nats-client==0.9.2',
        'jaeger-client==4.1.0',
        'fluent-logger==0.9.3',
        'Flask==1.1.1',
        'bottle==0.12.18',
        'coloredlogs==14.0',
        'colorama==0.4.0',
        'Pygments==2.6.1',
        'event-signal==1.8.0',
        "redis==2.10.6",
    ]

)



import os
import sys
import re
from .loggs import Loggs


class Options:
    App = None
    Debug = False
    Hosts = []
    Name = None
    Public = False
    Version = None
    Legacy = False
    Logger = Loggs('Options')
    isNats = False

    def __init__(self, app, name, version, hosts, is_nats=False):
        """
            Clase para la construccion de las opciones del microservicio.

            @params app Nombre de la aplicacion
            @params name Nombre del microservicios
            @params version Numero de la version del microservicio
            @params hosts Lista de direcciones de kafka o Nats.
            @params public Indicador si el servicio es de accesso publico

            @returns void
        """
        # Validar que se pase la direccion de kafka de forma correcta
        if not isinstance(hosts, list):
            self.Logger.error(
                'No se proporciono una lista de direcciones de kafka correcta')
            sys.exit(-1)
        elif len(hosts) < 1:
            self.Logger.error('Se proporciono una lista vacia de Kafka Hosts')

        # Asignar la variable debug a verdadero si no es productivo.
        self.Debug = False if os.environ.get('PRODUCTION', False) else True

        # Asignar el nombre de la aplicacion
        self.App = app

        # Asignar el nombre del microservicio
        self.Name = name

        # Asignar la version
        self.Version = version

        # Validar si es Nats
        if os.environ.get('NATS', None) is not None:
            self.isNats = True
        elif len(re.findall(r'^nats.*', hosts[0])) > 0:
            self.isNats = True
        else:
            self.isNats = False

        # Revalidar los brokers
        if len(os.environ.get('NATS', '')) > 0:
            self.Hosts = os.environ.get('NATS', '').split(',')
        elif len(os.environ.get('KAFKA', '')) > 0:
            self.Hosts = os.environ.get('KAFKA', '').split(',')
        else:
            # Asignar el cluster de kafka
            self.Hosts = hosts



import sys
import inspect
import json
import os
import socket
import copy
import base64
import datetime
import traceback
import uuid
import re
import _thread
import asyncio  # Requerido
from jaeger_client import Config
from opentracing import Format
from opentracing.ext import tags as ext_tags
from abc import abstractmethod, ABC
from .options import Options
from .Worker import Worker
from .Fork import Fork
from .ErrorMs import ErrorMs
import time
from .Util import validForksConf, getItemBy
from .version_framework import version
from http.server import BaseHTTPRequestHandler, HTTPServer
from .types import TypesActions
from .HttpResponse import HttpResponse
import redis
import hashlib

if os.environ.get('KAFKA', None):
    from .ms_base import KafkaBase as BaseClass
elif os.environ.get('NATS', None):
    from .NatsBase import NatsBase as BaseClass
else:
    # En caso que se no indique por variable de entorno
    from .ms_base import KafkaBase as BaseClass


class Microservices(BaseClass, ABC):
    """
        Clase con la logica necesaria para utilizar sobre los
        microservicios, se debe implementar los metodos segun
        como se requiera.
    """
    # Datos de configuracion
    __OPT = {}
    # Acciones del microservicio
    __ACTIONS = {}
    # Lista de errores definidos por el usuario
    __ERRORS = {
        '-97': 'No existe la accion que esta tratando de llamar',
        '-98': 'Ocurrio una error, favor de intentar de nuevo.',
        '-99': 'Ocurrio una excepcion, favor de intentar de nuevo.',
        '0': 'Smoke Test',
        '-1': 'Este servicio no puede ser consumido de forma externa'
    }
    # Logica de la aplicacion
    __APP = None

    # Topico
    __TOPIC__ = ''

    # Hostname
    __HOSTNAME = ''

    # Fork Configurados en la clase
    __FORKS = None

    # Indicar si es el servicio es una bifurcacion
    ___isForks = False

    # OpenTracing Jaeger
    ___Tracer = None

    # Indicar si se usara los logs de fluentd
    __fluentd_logs = False

    """
        Clase que contiene la logica del microservicio.
    """

    # Instancia de redis
    redisclient = None

    def __init__(self, opt):
        # Validar que se pase un objeto de configuracion correcto
        if not isinstance(opt, Options):
            self.logs.error('No se proporciono una configuracion correcta')
            sys.exit(-1)

        # Asignar los datos de la configuracion para su acceso
        self.__OPT = opt

        # Construccion del topico
        if opt.App == 'appcoppel' and opt.Version == 'v1':
            topico = opt.Name
        else:
            topico = "{}_{}_{}".format(opt.App, opt.Version, opt.Name)

        # Asignar el topico
        self.__TOPIC__ = topico

        # Recuperar el hostname del servidor
        self.__HOSTNAME = socket.gethostname()

        # Inicializacion del tracer
        self.initilizeTracer()

        # Inicializacion de HealthCheck
        _thread.start_new_thread(
            self.startHttpHealthCheck, (self.HealthCheck, self.smoketest))

        # Buscar las acciones
        self.__initActions()

        # Validar si es Productivo o si se activaron los logs a fluent
        if not os.environ.get('LOGS_FLUENT', None) is None:
            self.__fluentd_logs = True
            # Enviar la configuracion necesaria
            self.logs.set_conf(opt.App, opt.Name,
                               opt.Version, opt.Hosts, version)

        # Crear el administrador de la logica de la aplicacion.
        if self.__FORKS is None:
            self.__APP = Worker(self.__TOPIC__)
        else:
            self.logs.info(
                'Se Inicia el servicio con soporte para bifurcaciones')
            self.__APP = Fork(self.__TOPIC__)
            self.___isForks = True

        # Inicializar el redis
        self.initRedis()

        # llamar el constructor padre
        BaseClass.__init__(self, topico, opt.Hosts, topico)

        if opt.isNats:
            # iniciar la aplicacion
            """
                NOTA:
                    Ninguna instruccion debajo del etodo run_forever()
                    se ejecutara.
            """
            self.loop = asyncio.get_event_loop()
            self.loop.run_until_complete(asyncio.wait([self.run()]))
            # self.loop.run_until_complete(self.run())
            try:
                self.loop.run_forever()
            finally:
                self.loop.close()

    def initRedis(self):
        """
            Metodo para inicializar el redis
        """
        self.logs.info("Redis {}".format(
            os.environ.get('REDISVALIDATE', None)))

        if os.environ.get('REDISVALIDATE', None) is None:
            self.logs.warn('No se encuentra activado la validacion por redis')
        else:
            port = os.environ.get('REDIS_PORT', 6379)

            __pool_redis = redis.ConnectionPool(
                host=os.environ.get("REDISVALIDATE"), port=6379, db=0)
            self.redisclient = redis.Redis(connection_pool=__pool_redis)

            if self.redisclient is None:
                self.logs.warn(
                    'No fue posible conectarse al servidor \
                        de redis en la direccion: {}:{}'.format(
                        os.environ.get('REDISVALIDATE'), int(port)
                    ))
                sys.exit(-1)
            self.logs.info('Se encuentra activado la validacion por redis')

    def hasProcessMessage(self, data):
        """
            Metodo para validar que una operacion ya fue relizada.
        """
        if os.environ.get('REDISVALIDATE', None) is None:
            return False
        else:
            shahex = hashlib.sha224(bytes(data, 'utf-8')).hexdigest()
            response_redis = self.redisclient.getset(shahex, data)
            if response_redis is None:
                self.redisclient.expire(shahex, (10 * 60))
                return False
            return True

    def initilizeTracer(self):
        config = Config(
            config={
                'sampler': {
                    'type': 'const',
                    'param': 1,
                },
                'local_agent': {
                    'reporting_host': os.environ.get('JAEGER_AGENT_HOST', 'jaeger-agent.logging'),
                    'reporting_port': os.environ.get('JAEGER_AGENT_PORT', 6831),
                },
                'logging': True if os.environ.get('PRODUCTION', False) else False,
                'tags': {
                    '@coppel-py-version': version
                }
            },
            service_name=self.__TOPIC__,
            validate=True,
        )
        self.___Tracer = config.initialize_tracer()

    def send(self, topic, msj, idTransaction=""):
        """
            Metodo para el envio de datos a Nats

            @params reply Elemento al que se le responde
            @params msj Mensaje que enviara
            @params idTransaction Id de la transaccion
        """
        self._send(topic, msj, idTransaction)

    def HealthCheck(self):
        return self.validateConnection()

    def startHttpHealthCheck(self, healthcheck, smoketest):
        if os.environ.get('PRODUCTION', False):
            class myHandler(BaseHTTPRequestHandler):
                def do_GET(self):
                    message_parts = []
                    code = 0
                    try:
                        if healthcheck() and smoketest():
                            code = 200
                            message_parts.append("OK")
                    except Exception:
                        code = 500

                    if code < 100:
                        code = 500
                        message_parts.append("ERROR")

                    message = '\r\n'.join(message_parts)
                    self.send_response(code)
                    self.send_header(
                        'Content-Type', 'text/plain; charset=utf-8')
                    self.end_headers()
                    self.wfile.write(message.encode('utf-8'))

                def log_message(self, format, *args):
                    return

            self.logs.info('[HealthCheck] listening on 0.0.0.0:80')

            server = HTTPServer(('', 80), myHandler)
            server.serve_forever()
        else:
            self.logs.warn('HttpHealthCheck is Disabled')

    def coro_manager(self, coro):
        """
            Metodo para poder hacer el llamado de la funcion Async
            sin necesidad de usar await.

            @params coro Metodo async a llamar
        """
        try:
            coro.send(None)
        except StopIteration as stop:
            return stop.value

    def _send_error(self, data, id_transaction):
        data_error = {
            "id": id_transaction,
            "_id": id_transaction,
            "error": data,
            "servicio": self.__TOPIC__,
            "fecha": int(time.time() * 1000)
        }

        if isinstance(data, dict):
            if "error" in data and "_id" in data:
                data.update(
                    {"fecha": int(time.time() * 1000)})
                data_error = data

        self._send(
            os.environ.get('TOPIC_ERROR_CRITIC', 'Errores_criticos'),
            data_error,
            id_transaction
        )

    def json_to_b64(self, json_in):
        """
            Metodo para convertir un diccionaro al formato
            correcto de los mensajes.

            @param json_in Dict de entrada
        """
        return base64.b64encode(str.encode(json.dumps(json_in)))

    def without_response(self, topic, data={}, id_transaction=None, action=None):
        """
            Metodo para el envio de una peticion a un topico, sin esperar
            una respuesta del mismo.
        """

        if topic == "Errores_criticos":
            # self._send(topic, data, id_transaction)
            self._send_error(data, id_transaction)
        else:

            id_trans = id_transaction if id_transaction is None else str(
                uuid.uuid4())

            self._send(topic, {
                "metadata": {
                    "uuid": None,
                    "id_operacion": str(uuid.uuid4()),
                    "id_transaction": id_trans,
                    "intents": 1,
                    "callback": None,
                    "callbacks": [],
                    "owner": self.__TOPIC__,
                    "uowner": self.__TOPIC__,
                    "uworker": "{}_WITHOUT_RESPONSE".format(self.__TOPIC__),
                    "worker": self.__TOPIC__,
                    "asynchronous": False,
                    "mtype": "WITHOUT_RESPONSE",
                    "bifurcacion": False,
                    "time": "",
                    "from": socket.gethostname(),
                    "method": "GET" if action is None else action,
                    "deviceType": "",
                    "span": None,
                    "filters": None,
                },
                "data": data,
                "headers": {
                    "Asynchronous": False,
                    "Authorizacion": ""
                }
            }, id_trans)

    def _message(self, msg):
        """
            Metodo que se encagra de procesar todos los
            mensajes que llegas desde Kafka.
        """
        # metodo que se ejecutara
        mth = None

        # Recuperar los datos de otro modo regresar
        data = msg.get('data', {})

        # Recuperar la metadata
        meta = msg.get('metadata', {})

        # Configuracion del worker
        confForks = None

        # Parametros solicitados
        paramsConf = []

        # Recuperar el rootSpan
        # msg['metadata']['span'] if 'span' in msg['metadata'] else None
        rootSpan = None

        # Validar si es un smoktest
        if 'smoketest' in list(data):
            span = None

            if rootSpan is None:
                span = self.___Tracer.start_span(operation_name='smoketest')
            else:
                span = self.___Tracer.start_span(
                    operation_name='smoketest', child_of=rootSpan)

            span.set_tag(ext_tags.HTTP_METHOD, msg['metadata']['method'])

            # Ejecutar la funcion del smoketest
            if self.smoketest():
                # Todo esta bien
                self.__response(msg, 0, True, msg.get('uuid', None), span=span)
            else:
                # Algo salio mal
                self.__response(
                    msg, -1, True, msg.get('uuid', None), span=span)

            span.finish()
            return

        # Validacion de redis
        if self.hasProcessMessage('{}_{}_{}'.format(
            self.__TOPIC__,
            meta.get('id_transaction', ''),
            json.dumps(msg.get('data', {})
                       )
        )):
            span = None

            if rootSpan is None:
                span = self.___Tracer.start_span(
                    operation_name='replication')
            else:
                span = self.___Tracer.start_span(
                    operation_name='replication', child_of=rootSpan)

            if os.environ.get('DISABLED_LOGS_VALIDATE', None) is None:
                self.logs.error('Se recibio una operacion que ya se realizo')

            span.set_tag(ext_tags.ERROR, 'hasProcessMessage')

            span.finish()
            return True

        # Solo actualizar los datos si e el servicio original
        if self.__TOPIC__ == msg['metadata']['callback']:
            if not msg['metadata'].get('data_inicial', None) is None:
                # Recuperar el metodo original
                if not msg['metadata'].get('original_method', None) is None:
                    msg['metadata']['method'] = msg['metadata']['original_method']

                # Recuperar el metodo original
                if not msg['metadata'].get('data_inicial', None) is None:
                    msg['data'] = msg['metadata']['data_inicial']

        # Validar si se encuentra en modo Debug
        if self.__OPT.Debug and self.__fluentd_logs:
            _data_ = {}
            _data_.update({'owner': meta.get('owner')})
            _data_.update({'method': meta.get('method')})
            _data_.update({'headers': msg.get('headers', {})})
            _data_.update({'data': msg.get('data', {})})
            self.logs.info('ENTRADA {}'.format(json.dumps(_data_)))
        elif self.__OPT.Debug:
            self.logs.info(
                "\n ENTRADA: {}\n METODO: {}\n HEADERS: {}\n DATA: {}".format(
                    meta.get('owner'),
                    meta.get('method'),
                    json.dumps(msg.get('headers', {})),
                    msg.get('data', {})
                ))

        """ try:
            TIME = msg['metadata']['time']
            DATE = None
            IS_Z = TIME.find('Z') > -1
            currDate = datetime.datetime.utcnow()

            if IS_Z:
                include_mil = re.search(
                    r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{1,}z",
                    TIME
                )
                if include_mil:
                    DATE = datetime.datetime.strptime(
                        TIME,
                        '%Y-%m-%dT%H:%M:%S.%fZ'
                    )
                else:
                    DATE = datetime.datetime.strptime(
                        TIME,
                        '%Y-%m-%dT%H:%M:%SZ'
                    )

                elapsedTime = currDate - DATE
                diff = divmod(elapsedTime.total_seconds(), 60)[0]
                if diff >= 5:
                    # print('EXIT TIME : ', msg['metadata']['time'], ' ', diff)
                    pass

        except Exception as e:
            self.logs.warn(
                'Error al momento de procesar el tiempo del mensaje', e) """

        # Procesar el mensaje
        try:
            mth = self.__getMethod(msg.get('metadata'))
        except Exception as identifier:
            # No fue posible recuperar una accion para el evento
            self.logs.error(
                'No fue posible recuperar una accion para el evento {}'.format(
                    identifier
                )
            )

        # Validar que exista
        try:
            if hasattr(mth, '__CONF_FORKS__'):
                methodConfFork = getattr(mth, '__CONF_FORKS__')
                # Recuperar la configuracion
                confForks = self.__getConfServiceFork(methodConfFork)
        except Exception as e:
            self.logs.error(
                'Ocurrio un error al recuperar la configuracion del servicio: {}'.format(e))

        # Buscar los parametros
        try:
            if hasattr(mth, '__CONF_PARAMS__'):
                paramsConf = getattr(mth, '__CONF_PARAMS__')
        except Exception as e:
            self.logs.error(e)

        if rootSpan is not None:
            rootSpan = self.___Tracer.extract(Format.TEXT_MAP, rootSpan)

        span = self.___Tracer.start_span(operation_name=getattr(
            mth, '__FUNC_NAME__'), child_of=rootSpan)

        span.set_tag(ext_tags.HTTP_METHOD, msg['metadata']['method'])

        RESP = {}

        # Enviar el mensaje al que procesa [Worker, Fork]
        try:
            # Configuracion de los forks
            if confForks is None:
                self.__APP = Worker(self.__TOPIC__)
            else:
                self.__APP = Fork(self.__TOPIC__)
                self.__APP.confForks(confForks)

            # Indicar si es una bifurcacion
            msg['metadata']['bifurcacion'] = True if confForks else False

            # ejecutar el proceso
            RESP = self.__APP.process(span, msg, mth, paramsConf)

            # Ruta [SOLO NATST]
            id_route_nats = getItemBy('response.uuid', RESP)

            if not msg['metadata']['mtype'] == 'WITHOUT_RESPONSE':
                RESPONSE = RESP.get('response', {})
                ERROR_CODE = RESP.get('errorCode')
                ERROR_WORKER = RESP.get('errorWorker', None)
                # Enviar la respuesta
                self.__response(
                    RESPONSE,
                    ERROR_CODE,
                    span=span,
                    nats_route=id_route_nats,
                    errorResp=ERROR_WORKER,
                )

        except ErrorMs as error:
            span.set_tag(ext_tags.ERROR, True)

            # Recuperar el registro del error
            error_reg = self.___getErrorId(error.errorCode)

            # Validar si es un diccionario
            isDict = isinstance(error_reg, dict)

            span.log_kv({
                'event': 'error',
                'error.code': error.errorCode,
                'error.message': error_reg.get(
                    'message', ""
                ) if isDict else error_reg,
                'error.stack': traceback.format_exc()
            })

            self.logs.error(error_reg.get(
                'message', ""
            ) if isDict else error_reg)

            # Verificar si es un codigo de error ya especificado
            if error.http_code is not None:
                msg["response"]["data"]["response"] = HttpResponse(
                    msg["response"]["data"]["response"],
                    error.http_code,
                    error.message,
                    error.headers
                )
            else:
                if isDict and "http_code" in error_reg:
                    msg["response"]["data"]["response"] = HttpResponse(
                        msg["response"]["data"]["response"],
                        error_reg["http_code"],
                        error_reg.get('message', ""),
                        error.headers
                    )

            # Validar si es async
            if not msg['metadata']['mtype'] == 'WITHOUT_RESPONSE':
                self.__response(
                    msg, error.errorCode, span=span, nats_route=msg.get(
                        'uuid', None), errorResp=RESP.get('errorWorker', None),
                )  # Enviar el error

        except Exception as e:
            span.set_tag(ext_tags.ERROR, True)
            span.log_kv({'event': 'error', 'error.message': str(
                e), 'error.stack': traceback.format_exc()})
            self.logs.error(e)
            # Enviar el error
            if not msg['metadata']['mtype'] == 'WITHOUT_RESPONSE':
                self.__response(
                    msg, -99, span=span, nats_route=msg.get('uuid', None),
                    errorResp=RESP.get('errorWorker', None),
                )

        span.finish()

    def __response(self, data, errorCode, isSmokeTest=False,
                   nats_route=None, span=None, errorResp=None):
        """
            Metodo para enviar la respuesta a kafka
        """

        # Diccionario de la respuesta
        Resp = {}

        # Datos http.
        HTTP = None

        # Respuesta del servicio
        RESPONSE_DATA = None if data.get('response', None) is None else \
            data["response"]["data"]["response"]

        # Validar si data no es de tipo httpResponse
        if (isinstance(RESPONSE_DATA, HttpResponse)):
            http_data = data["response"]["data"]["response"]
            HTTP = {
                "code": http_data.code,
                "headers": http_data.headers,
                "message": http_data.message
            }
            # Reset de data
            data["response"]["data"]["response"] = http_data.data

        # Crear una copia de la respuesta
        dataResponse = data

        # Topico de respuesta
        TOPIC_RESP = None

        # Codigo de error
        errorCodeBk = 0

        # Comprobar si contiene un error heredado
        if errorResp is not None:
            errorCode = errorResp.get('errorCode', -99)  # Error generico

        try:
            errorCodeBk = errorCode
            # Prevenir el paso de string como error
            errorCode = int(errorCode)
        except Exception:
            self.logs.error(
                'Ocurrio une error al con el CAST del error code: {}'.format(
                    errorCodeBk
                )
            )
            errorCode = -99

        # Validar si ocurrio un error
        if errorCode < 0 or errorCode > 0:
            # Mensaje del error
            message_error = ""

            # Registro del error
            reg_error = self.___getErrorId(errorCode)

            # Validar si es un string o un objeto de error
            if isinstance(reg_error, dict):
                message_error = reg_error.get('message', "")
            else:
                message_error = reg_error

            # Estructura del mensaje de error
            dataResponse['response'] = {
                "data": {
                    "response": {
                        "hostname": self.__HOSTNAME,
                        "code": errorCode,
                        "errorCode": errorCode,
                        # Regresar el mensaje que se proporciono desde el
                        # worker que se consumio
                        "userMessage": message_error if errorResp is None else errorResp.get('userMessage')
                    }
                },
                "meta": {
                    "id_transaction": dataResponse['metadata']['id_transaction'],
                    "status": 'ERROR' if errorCode < 0 or errorCode > 0 else 'SUCCESS'
                }
            }

        # Asignar el tiempo
        dataResponse['metadata']['time'] = datetime.datetime.utcnow().isoformat()

        # Indicar el worker
        # dataResponse['metadata']['worker'] = data['metadata']['owner']

        # Topico
        dataResponse['metadata']['owner'] = self.__TOPIC__

        # Tipo de salida
        dataResponse['metadata']['mtype'] = 'output'

        # Indicar Hostname
        dataResponse['metadata']['hostname'] = self.__HOSTNAME

        # Indicar version del framework
        dataResponse['metadata']['framework'] = version

        # Validar que tenga un uowner
        if dataResponse['metadata'].get("uowner", None) is not None:
            dataResponse['metadata']['uworker'] = data['metadata']['uowner']

        # Validar si tiene asignado un uworker
        if dataResponse['metadata'].get("uworker", None) is not None:
            dataResponse['metadata']['uowner'] = data['metadata']['uworker']

        if span:
            carrier = {}
            self.___Tracer.inject(span, Format.TEXT_MAP, carrier)
            dataResponse['metadata']['span'] = carrier

        # Ver si es un Worker
        if isinstance(self.__APP, Worker):
            dataResponse['metadata']['bifurcacion'] = False

        # Ver si es bifurcacion
        is_forks = getItemBy('metadata.bifurcacion', dataResponse)

        # Metodo original
        original_method = getItemBy('metadata.original_method', dataResponse)

        '''
            Validacion del Callback
        '''

        if is_forks:
            if getItemBy('metadata.callbacks', dataResponse) is None:
                # Agregar
                dataResponse['metadata'].update({'callbacks': []})
                # Validar que sea NATS
                if nats_route is not None or len(nats_route) > 0:
                    # Agregar el Nats_UUID
                    dataResponse['metadata']['callbacks'].append({
                        "method": 'ENDPOINT',
                        "name": nats_route,
                    })

            # Recuperar los callbacks
            callbacks = getItemBy('metadata.callbacks',
                                  copy.deepcopy(dataResponse))

            if len(callbacks) > 0:
                cllbs = callbacks.pop()

                tp = cllbs.get('name')
                mt = cllbs.get('method')

                if self.__TOPIC__ == tp and original_method == mt:
                    self.logs.warn('Repetition event prevented')
                else:
                    # Agregar al arreglo
                    dataResponse['metadata']['callbacks'].append({
                        "method": original_method,
                        "name": self.__TOPIC__,
                    })
            else:
                # Agregar al arreglo
                dataResponse['metadata']['callbacks'].append({
                    "method": original_method,
                    "name": self.__TOPIC__,
                })

            # TOPIC_RESP = dataResponse['metadata']['worker']
        else:
            callbacks = getItemBy('metadata.callbacks', dataResponse)
            # Se obtiene el ultimo elemento agregado para responder a el
            if callbacks and len(callbacks) > 0:
                cllbs = callbacks.pop()
                sizeCllbs = len(callbacks)
                mthd = original_method
                find = True
                counter = 0
                if self.__TOPIC__ == cllbs.get('name') and mthd == cllbs.get('method'):
                    while find and counter < sizeCllbs:
                        cllbs = callbacks.pop()
                        if self.__TOPIC__ != cllbs.get('name') or mthd != cllbs.get('method'):
                            TOPIC_RESP = cllbs.get('name')
                            dataResponse['metadata']['method'] = cllbs.get(
                                'method')
                            find = False
                        counter += 1
                else:
                    TOPIC_RESP = cllbs.get('name')
                    dataResponse['metadata']['method'] = cllbs.get('method')

            else:
                pass

            original_method = dataResponse['metadata'].get(
                'original_method', None)

            # Eliminamos elemento (Este es agregado cuando es una bifurcacion)
            if original_method is not None:
                del dataResponse['metadata']['original_method']
            else:
                pass

        # Ver si es utilizado Nats
        if nats_route is None:
            OWNER = dataResponse['metadata']['owner']
            # Topico de respuesta
            TOPIC_RESP = TOPIC_RESP if TOPIC_RESP else "respuesta_{}".format(
                OWNER)
        else:
            # Validar si existe un topico seleccionado
            if TOPIC_RESP is None:
                TOPIC_RESP = nats_route
            else:
                # Comprobar si corresponde a un topico de servicio o sistema
                is_topic = re.search(
                    "[a-zA-Z0-9]{1,}[_.]{1}[a-zA-Z0-9]{1,}[_.]{1}.*",
                    TOPIC_RESP
                )
                # Asignar el id de ruta que proviene del reply
                TOPIC_RESP = TOPIC_RESP if is_topic else nats_route

        # Consultar quien es el callback
        cllb = getItemBy('metadata.callback', dataResponse)

        # Datos requeridos para seleccion de topic
        req_worker = getItemBy('metadata.worker', dataResponse)
        req_owner = getItemBy('metadata.owner', dataResponse)
        req_bifurcacion = getItemBy('metadata.bifurcacion', dataResponse)

        if req_worker != req_owner and req_bifurcacion:
            TOPIC_RESP = req_worker
            dataResponse['metadata']['worker'] = req_owner
            Resp = dataResponse
        elif self.__TOPIC__ != cllb and req_bifurcacion:
            TOPIC_RESP = TOPIC_RESP if TOPIC_RESP else cllb
            dataResponse['data'] = dataResponse['response']['data']
            Resp = dataResponse
        else:
            # Asignar la respuesta
            Resp = {
                "_id": dataResponse['metadata']['id_transaction'],
                "data": getItemBy('metadata.data_inicial', dataResponse),
                "headers": dataResponse.get('headers', {}),
                "metadata": dataResponse.get('metadata', {}),
                "response": dataResponse.get('response', {})
            }

        callbacksTop = getItemBy('metadata.callbacks', dataResponse)

        if callbacksTop and len(callbacksTop) == 0:
            del Resp['metadata']['metadata_inicial']
            del Resp['metadata']['data_inicial']

        id_transaction = dataResponse['metadata']['id_transaction']

        # Revisar si es un smoketest
        if isSmokeTest:
            # respuesta generica
            Resp.update(
                {
                    'response': {
                        "data": {
                            "response": {
                                "code": errorCode,
                                "framework": version,
                                "hostname": self.__HOSTNAME,
                                "userMessage": "Smoke test"
                            }
                        },
                        "meta": {
                            "id_transaction": id_transaction,
                            "status": 'ERROR' if errorCode < 0 or errorCode > 0 else 'SUCCESS'
                        }
                    }
                }
            )

        span.log_kv({'event': 'output', 'response': Resp['response']})

        if nats_route is not None:
            Resp.update({"uuid": nats_route})

        # Validar si se envia el objeto HTTP perzonalizado
        if HTTP is not None:
            Resp['metadata'].update({
                "http": HTTP
            })
        else:
            # Datos por default
            Resp['metadata'].update({
                "http": {
                    "code": 500 if errorCode < 0 or errorCode > 0 else 200,
                    "headers": {},
                    "message": ""
                }
            })

        # Validar si es modeo DEBUG
        if self.__OPT.Debug and self.__fluentd_logs:
            self.logs.info("SALIDA [{}]".format(TOPIC_RESP), Resp)
        elif self.__OPT.Debug:
            self.logs.info("SALIDA [{}]: {}".format(
                TOPIC_RESP, json.dumps(Resp)))

        # Publicar la respuesta
        self._send(TOPIC_RESP, Resp,
                   dataResponse['metadata']['id_transaction'])

    def ___getErrorId(self, id_error):
        """
            Metodo apra recuperar los datos de un error registrado
        """
        try:
            if self.__ERRORS.get(str(id_error), None):
                return self.__ERRORS.get(str(id_error))
            else:
                return self.__ERRORS.get('-99')
        except Exception:
            self.logs.error(
                'Ocurrio un excepcion al recuperar los datos del error')
            return 'Ocurrio un error generico'

    def __getListener(self):
        """
            Metodo para recuperar la accion registrada
            en el Listener, de no contar con dicha accion
            se lanza una excepcion.
        """
        if self.__ACTIONS.get(TypesActions.LISTENER, None):
            return self.__caller(self.__ACTIONS.get(TypesActions.LISTENER))
        else:
            raise Exception('No existe un metodo para el evento solicitado')

    def __getMethod(self, meta):
        """
            Metodo para recuperar el metodo que se
            ejecutara, segun los datos pasados en
            el metada.
        """
        # Recuperar el metodo
        METHOD = meta.get('method', None)

        # Metodo que se ejecutara
        FNC = None

        # Seleccion del metodo
        if METHOD:
            if (meta.get('uuid', None) and len(meta.get('uuid', '')) > 0):
                if METHOD == 'GET':  # Consultar un elemento
                    FNC = self.__caller(self.__ACTIONS.get(TypesActions.GET))
                elif METHOD == 'DELETE':  # Eliminar el elemento
                    FNC = self.__caller(
                        self.__ACTIONS.get(TypesActions.DELETE))
                elif METHOD == 'PUT':  # Actualizar el elemento
                    FNC = self.__caller(
                        self.__ACTIONS.get(TypesActions.UPDATE))
            elif METHOD == 'GET':  # Lista de servicios
                if self.__ACTIONS.get(TypesActions.LIST, None):
                    FNC = self.__caller(self.__ACTIONS.get(TypesActions.LIST))
            elif METHOD == 'POST':  # Creacion de un nuevo elemento
                if self.__ACTIONS.get(TypesActions.CREATE, None):
                    FNC = self.__caller(
                        self.__ACTIONS.get(TypesActions.CREATE))

        # Validar si se recuper un metodo asignado
        if FNC is not None:
            return FNC
        else:
            # Regresar el Listener por default
            return self.__getListener()

    def __initActions(self):
        """
            Metodo que se encarga de recuperar todas las
            acciones registradas, para su implementacion
            durante su llamado.
        """
        for f in inspect.getmembers(self):
            # Validar que tenga el atributo minimo
            if hasattr(f[1], '__MICROSERVICE_ACTION__'):
                # Recuperar el tipo de accion
                typeAction = getattr(f[1], '__TYPE__')
                # Validar si es el de errores
                if typeAction == TypesActions.ERRORS:
                    # Recuperar la funcion para ejecutarla
                    errorFNC = self.__caller(f[0])
                    # Ejecutar la funcion para recuperar los errores definidos
                    errores_definidos = errorFNC()
                    # Validar que sea el tipo correcto
                    if not isinstance(errores_definidos, dict):
                        self.logs.error(
                            'No se proporciono un formato correcto para los errores definidos')
                        sys.exit(-1)
                    # Ejecutar y asignar los errores
                    self.__ERRORS = self.__merge_dicts(
                        self.__ERRORS, errores_definidos)
                elif typeAction == TypesActions.FORKS:
                    # Recuperar la funcion para ejecutarla
                    forksFNC = self.__caller(f[0])
                    # Ejecutar la funcion para recuperar la configuracion
                    forks_conf = forksFNC()
                    # Validar que sea el tipo correcto
                    if not isinstance(forks_conf, dict):
                        self.logs.error(
                            'No se proporciono un formato correcto para los Forks definidos')
                        sys.exit(-1)
                    # Validar que cuente con el forma correcto
                    valid, error = validForksConf(forks_conf)
                    if valid:
                        self.__FORKS = forks_conf
                    else:
                        self.logs.error(error)
                        sys.exit(-1)
                else:
                    # Almacenar la accion
                    self.__ACTIONS.update({typeAction: f[0]})

    def __caller(self, name):
        """
            Metodo para recuperar una propiedad que sera utilizada
            como metodo
        """
        if hasattr(self, name):
            return getattr(self, name)

    @abstractmethod
    def smoketest(self):
        """
            Metodo que es llamado para validar los
            servicios desde su consumo por REST/Kafka
        """
        pass

    def __merge_dicts(self, *dict_args):
        """
        Given any number of dicts, shallow copy and merge into a new dict,
        precedence goes to key value pairs in latter dicts.
        """
        result = {}
        for dictionary in dict_args:
            result.update(dictionary)
        return result

    def __getConfServiceFork(self, conf):
        """
            Metodo para iniciar la configuracion de los Forks a ejecutar
        """
        forks = []

        # Recorrer las configuraciones
        for item in conf:
            # Verificar que existe en las configuraciones
            if self.__FORKS.get(item.get('fork'), None) is None:
                self.logs.error(
                    'No se encuentra configurado el Fork {}'.format(item.get('fork')))
                raise Exception(
                    'No se encuentra configurado el Fork {}'.format(item.get('fork')))

            # Actualizar el item con la configuracion inicial
            item.update({"conf": self.__FORKS.get(item.get('fork', ''), None)})

            # Validar si es necesario recuperar una funcion
            if not item.get('fnc', None) is None:
                item.update({'fnc_call': self.__caller(item.get('fnc'))})

            # Agregar a la coleccion
            forks.append(item)

        # Regresar la coleccion
        return forks
